︠80d8d986-5f59-41d0-a4cc-4bad06eae51ei︠
%html
<p style="text-align: center;"><strong><span style="font-size: large;"><span style="font-family: arial, helvetica, sans-serif;">Colormap for implicit_plot3d and parametric_plot3d (#12212)</span></span></strong></p>
<p style="text-align: center;"><strong>jvkersch</strong></p>
<p style="text-align: left;">The class below is based on some code from <em>sage.plot.plot3d.plot3d </em>and allows for a texture attribute of a surface to be varied gradually. &nbsp;The precise way in which the attribute varies can be specified by means of a scalar function, the most straightforward ones being projection onto the x, y, or z-axis. &nbsp;In this way, it is possible to apply a colormap to a surface, to create partially translucent objects, or to color in surfaces in a variety of ways (for instance, to visualize Gaussian curvature).</p>

︡8ea36ba3-64ec-4c55-b751-f9a7f74e0fc7︡{"html": "<p style=\"text-align: center;\"><strong><span style=\"font-size: large;\"><span style=\"font-family: arial, helvetica, sans-serif;\">Colormap for implicit_plot3d and parametric_plot3d (#12212)</span></span></strong></p>\n<p style=\"text-align: center;\"><strong>jvkersch</strong></p>\n<p style=\"text-align: left;\">The class below is based on some code from <em>sage.plot.plot3d.plot3d </em>and allows for a texture attribute of a surface to be varied gradually. &nbsp;The precise way in which the attribute varies can be specified by means of a scalar function, the most straightforward ones being projection onto the x, y, or z-axis. &nbsp;In this way, it is possible to apply a colormap to a surface, to create partially translucent objects, or to color in surfaces in a variety of ways (for instance, to visualize Gaussian curvature).</p>"}︡
︠834a0d19-062b-43bb-a7f4-3e19cb721e5b︠
from sage.plot.plot3d.base import Graphics3dGroup
from sage.structure.sage_object import SageObject

class GradualTextureTransform(SageObject):
    def __init__(self, attribute, values, fun=None, bounds=None):
        """
        INPUT::
            
            - ``attribute`` -- texture attribute to transform.  
            If this is not one of ``color``, ``opacity``, 
            ``ambient``, ``diffuse``, ``specular`` or ``shininess``,
            the transformation has no effect.
           
            - ``value_range`` -- range of values over which the 
            attribute should range.  In the case where the color 
            attribute is transformed, this can be a colormap.
            
            - ``fun`` -- scalar-valued function specifying the 
            change in attribute.  If this keyword is omitted, 
            the projection x, y, z -> z will be used.  This 
            could be either a symbolic expression, or one of 
            the following: 'height_x', 'height_y', 'height_z', 
            denoting respectively the height above the x-, y-, 
            or z-plane.
            
            - ``bounds`` -- tuple representing the range for the 
            scalar function ``fun``.  If this keyword is omitted, 
            some preset defaults will be used.    
            
        """

        self.attribute = attribute
        self.fun       = fun
        self.bounds    = bounds
        self.values    = list(values)

        self.projection_axis = 2

        if fun is None:
            self.fun = lambda x, y, z: z
            self.projection_axis = 2
        elif isinstance(fun, str):
            if fun is "height_x":
                self.fun = lambda x, y, z: x
                self.projection_axis = 0
            elif fun is "height_y":
                self.fun = lambda x, y, z: y
                self.projection_axis = 1
            elif fun is "height_z":
                self.fun = lambda x, y, z: z
                self.projection_axis = 2
            else:
                raise ValueError, "Not a valid texture function:", fun


    def apply(self, group):
        """
        Apply this texture transformation to a graphical object.
        
        INPUT::
            
            - ``group`` -- a graphical object (``Graphics3d``) or 
            group of objects (``Graphics3dGroup``).
            
        OUTPUT::
            
            - A ``Graphics3dGroup`` representing the transformed object.     
        """

        if not isinstance(group, Graphics3dGroup):
            group = Graphics3dGroup([group])

        # Code below adapted from plot3d_adaptive
        if self.bounds is None:
            # Determine limits for function from bounding box
            bounds = group.bounding_box()

            min_z = bounds[0][self.projection_axis]
            max_z = bounds[1][self.projection_axis]
        else:
            min_z = self.bounds[0]
            max_z = self.bounds[1]

        if max_z == min_z:
            span = 0
        else:
            span = (len(self.values)-1) / (max_z - min_z)  

        def normalized_fun(x, y, z):
            val = self.fun(x, y, z) 
            if val > max_z: val = max_z
            if val < min_z: val = min_z
            return int((val-min_z)*span)

        new_all = []
        kwds = {}
        for obj in group.all:
            obj.triangulate()  # Faces have to exist before object can be partitioned
            parts = obj.partition(normalized_fun) 
     
            for k, G in parts.iteritems():
                kwds[self.attribute] = self.values[k]
                G.set_texture(kwds)
                new_all.append(G)

        return Graphics3dGroup(new_all)



class ColormapTransform(GradualTextureTransform):
    """
    Transformation to apply a colormap to a given surface.  
    
    See ``GradualTextureTransform`` for more details.
    """
    def __init__(self, cmap, fun=None, bounds=None):
        GradualTextureTransform.__init__(self, 'color', cmap, fun, bounds)
        

class OpacityTransform(GradualTextureTransform):
    """
    Transformation to change the opacity of given surface in a non-uniform way. 
    
    See ``GradualTextureTransform`` for more details.
    """

    def __init__(self, values, fun=None, bounds=None):
        GradualTextureTransform.__init__(self, 'opacity', values, fun, bounds)
︡c4b21603-86fe-4c05-a52c-e4c74c5c53a0︡︡
︠4daab4d1-7325-445d-952e-57d58199f73c︠
# Predefined colormaps
cmsel1 = [colormaps['autumn'](i) for i in sxrange(0,1,0.05)]
cmsel2 = [colormaps['spectral'](i) for i in sxrange(0,1,0.02)]
︡be48f8bc-3d73-4c83-a7a2-c507a0202c2f︡︡
︠3fe5d7da-ea30-48d9-ab5a-d169bbb2c3d7︠
# The original example from the ticket, where using a colormap worked out of the box
︡14bde78d-0df5-412f-b623-e7fdf0309e45︡︡
︠fec819a9-ae4b-42a9-911c-37196638d8c1︠
var('r v')
p = plot3d(0.2*(r**2 + v**2) + cos(2*r)*sin(2*v),(r,-2,2), (v,-2,2), adaptive=True, color=cmsel1, plot_points=10, opacity=0.9)
p2 = sphere((0,0,0),1,color='black',opacity=0.5)
(p+p2).show(aspect_ratio=(1,1,1))
︡0ac1a1fc-1fa9-4572-bf11-efda3eafa2db︡︡
︠c88ed9ab-e975-4c59-a973-a7ae6bb805b6︠
# The same example, but with the functionality of the ColormapTransform class:
︡fded4e61-5307-442d-b649-f4b66ab3655e︡︡
︠df80d7e2-6ad6-44b4-9f84-d5d8fc89ceed︠
var('r v')
c = ColormapTransform(cmsel1)
p1 = plot3d(0.2*(r**2 + v**2) + cos(2*r)*sin(2*v),(r,-2,2), (v,-2,2), opacity=0.9)
p2 = sphere((0,0,0),1,color='black',opacity=0.5)
p3 = c.apply(p1)
(p2+p3).show(aspect_ratio=(1,1,1))
︡73390b57-3ffd-438b-b0de-b905aa00c174︡︡
︠112a7b83-f9f9-4e63-93a0-0bff81c29b39︠
# An example of an implicit surface with a color map (this is the non-functioning example from the ticket):
︡5cd6318c-6ac4-485f-be45-352aee0e9f33︡︡
︠6c2d9d3f-fb37-4c68-be81-e2f574c3e8ee︠
c = ColormapTransform(cmsel1)
var('x,y,z')
p1 = implicit_plot3d(x^2+y^2+z^2==4, (x, -3, 3), (y, -3,3), (z, -3,3))
p2 = c.apply(p1)
p2.show()
︡b6b84ab1-ce5d-4b8e-8287-50bcea34bd5f︡︡
︠893a07a2-879e-491e-8431-dfa513d253c0︠
# By default, the color values are selected according to the z-values of the points on the surface.  Other functions are possible, too:
︡e202629a-c78e-495b-a396-2f47d2199c5b︡︡
︠63ad8723-47f7-4a6a-b652-c9a583b74b87︠
# Uniform gradient along the x-axis
var('r v')
p = plot3d(0.2*(r**2 + v**2) + cos(2*r)*sin(2*v),(r,-2,2), (v,-2,2))
c = ColormapTransform(cmsel1, 'height_x')
c.apply(p).show(aspect_ratio=(1,1,1))
︡34bb154b-9498-4806-8e4c-0d07b8774f94︡︡
︠5d575d59-e94f-4915-bbe3-655400bef278︠
# Non-uniform coloring.  Here a function is specified.
︡a3a4afe1-b8fd-41e0-8b64-40b88fed440c︡︡
︠6527e6ef-dda3-4c19-85bc-1da77e503867︠
var('x,y,z')
p = plot3d(0.2*(r**2 + v**2) + cos(2*r)*sin(2*v),(r,-2,2), (v,-2,2))
c = ColormapTransform(fun=lambda x, y, z: x+y, bounds=(-4, 4), cmap=cmsel1)
c.apply(p)
︡1056c6f0-78ad-45cb-9870-4486b786a791︡︡
︠22b55557-8581-48be-b500-8686eefde4e5︠
# Another non-uniform coloring, this time with a nonlinear function (cos)
︡96f09204-5e36-4672-be49-e73fb5d0fa15︡︡
︠aa71c2b2-48ba-4188-9b73-fa354b59d583︠
var('r v')
p = implicit_plot3d(x^2+y^2+z^2==4, (x, -3, 3), (y, -3,3), (z, -3,3))
c = ColormapTransform(cmsel1, fun=lambda x, y, z: cos(2*pi*z/3-pi/3), bounds=(-1, 1))
c.apply(p).show(aspect_ratio=(1,1,1))
︡6f733d4e-6bd4-402a-9e7e-7aa442a048f5︡︡
︠bebc3b18-893b-41c2-9cd6-9cf764c24c7a︠
# A more complicated surface (adapted from the implicit_plot docs)
︡f51c6dde-2c42-4166-8d01-e05082730c5a︡︡
︠ce2dc0b8-c705-4307-8a35-7a142bd9363e︠
u, v = var('u,v')
fx = u -u^3/3  + u*v^2
fy = v -v^3/3  + v*u^2
fz = u^2 - v^2
p = parametric_plot3d([fx, fy, fz], (u, -2, 2), (v, -2, 2))
c = ColormapTransform(cmsel2)
c.apply(p)
︡f8c2e0aa-75ed-4a14-a110-22ee066b93ab︡︡
︠9052ba31-76f3-4c90-b46d-6e8869174a07︠
# We can also gradually vary other texture attributes, such as opacity:
# NOTE: I have no idea why the outer sphere is so coarse after the transformation is applied.  Increasing plot_points in the definition for p2 does not seem to help
    
p1 = sphere((0,0,0), .5, color='red')
p2 = sphere((0,0,0),  1)

opacity_values = sxrange(0, 1, .05)
c = OpacityTransform(opacity_values)
p3 = c.apply(p2)
(p1 + p3).show()
︡7354c218-d7e1-4b18-a7a6-c188e7c44ddd︡︡
︠9e0a79bc-2ea7-47f8-977e-8867a0638596︠
# The next three examples show implicit surfaces colored by Gaussian curvature
︡15589735-039e-4300-a52b-ef86b7f893d9︡︡
︠41c40508-0604-4b02-92ab-bb56e270dfc7︠
# Quick 'n' dirty routine to compute the Gaussian curvature of an implicit surface:
def gaussian_curvature_implicit_surface(f):    
    fx = diff(f, x); fy = diff(f, y); fz = diff(f, z)
    fxx = diff(fx, x); fxy = diff(fx, y); fxz = diff(fx, z)
    fyy = diff(fy, y); fyz = diff(fy, z); fzz = diff(fz, z)

    T1 = (fz*(fxx*fz - 2*fx*fxz) + fx^2*fzz).simplify_full()
    T2 = (fz*(fyy*fz - 2*fy*fyz) + fy^2*fzz).simplify_full()
    T3 = (fz*(-fx*fyz + fxy*fz - fxz*fy) + fx*fy*fzz).simplify_full()
    T4 = (fz^2*(fx^2 + fy^2 + fz^2)^2).simplify_full()

    K = ((T1*T2 - T3^2)/T4).simplify_full()
    return fast_float(K, 'x', 'y', 'z')
︡0df0bb12-6e44-49e1-863c-1f481983ffeb︡︡
︠4ad6bc79-abd8-4213-9a30-d44a17e839c6︠
# Ellipsoid 
f = x^2/4 + y^2 + z^2 - 1
p = implicit_plot3d(f, (x, -2, 2), (y, -1, 1), (z, -1, 1)) 
K = gaussian_curvature_implicit_surface(f)
︡b01ff0cd-d46d-4a3a-ac22-c84c7148486a︡︡
︠d5565f97-f7d8-4d5c-9466-7faa687aa26f︠
c = ColormapTransform(cmsel1, fun=K)
c.apply(p).show(aspect_ratio=(1, 1, 1))
︡34b94eb8-bb40-4a88-812b-1ba4662022b5︡︡
︠5822f0fb-31c2-46db-824f-f63e4d393b3e︠
# The previous picture suggest that the highest curvature can be found at the edges of the ellipsoid, whereas the lowest values for the curvature are in the center:
︡fd0ac6f0-687c-4994-b74a-7018625974f7︡︡
︠2b476945-3d2a-40c6-96c1-08da9e0a7b85︠
K_high = K(2, 0, 0); K_high
︡7253f9d2-92b1-41db-8b36-eee730023907︡{"stdout": "4.0"}︡
︠ce96e4a2-93c8-4763-b54c-f98502bd4bfc︠
K_low = K(0, 1, 0); K_low
︡85244e6a-118f-4795-8ec3-c8e626434e2b︡{"stdout": "0.25"}︡
︠6bb7e1b9-7e72-4e90-8ead-5fa18fd230a2︠
# Taking these bounds into account, we can produce a more accurate curvature plot:
c = ColormapTransform(cmsel1, fun=K, bounds=(K_low, K_high))
c.apply(p).show(aspect_ratio=(1, 1, 1))
︡434e3a29-c7a6-46a1-9bcd-1036293b45ef︡︡
︠4660167d-08bc-4dac-b2f2-c24ac3723e67︠
# This seems to suggest that the Gaussian curvature is quite low over most of the ellipsoid, and increases pretty quickly at the end points.  Let's try to verify this in a more pedestrian way: we take the intersection of the ellipsoid with the positive y-plane, which gives a curve, we compute the Gaussian curvature along this curve, and we plot this.

from numpy import linspace
points = []
for cx in linspace(-2, 2, 20):
    cy = sqrt(1-cx^2/4)
    K_val = K(cx, cy, 0)
    points.append([cx, K_val])
︡05b4f6a6-cb8b-4182-b2e7-ad097ae99cc1︡︡
︠f988cd75-230e-445f-9667-8adde4fe1f9e︠
# The curvature indeed increases quickly at the end points
list_plot(points, plotjoined=True)
︡ef91c481-98f1-4b41-9913-52297c4e8b34︡{"html": "<font color='black'><img src='cell://sage0.png'></font>"}︡
︠7422e0aa-db73-4116-ac7f-59864d9a57ef︠
var('x, y, z')
︡f037a101-d15b-459f-9cf3-34207899125d︡{"stdout": "(x, y, z)"}︡
︠139353bf-3546-4ac2-840f-4b8d5a6ecb3e︠
# A saddle surface (negative curvature at the origin)
f = z+x^2-y^2
p = implicit_plot3d(f, (x, -2, 2), (y, -2, 2), (z, -2, 2), plot_points=60, contour=0)
︡32da0b79-26f6-4e4d-a69b-ede246d2c119︡︡
︠c4041efd-06d8-4301-b131-20158d56b9e5︠
f
︡7bb1971c-9dc7-4711-b0ce-36943c5325c7︡{"stdout": "x^2 - y^2 + z"}︡
︠59801cf8-0128-4ca6-badf-7c3fcbc48a06︠
K = gaussian_curvature_implicit_surface(f)
︡f4054e49-f852-49d9-be05-d85252012a9d︡︡
︠ebaed2bd-a93a-4505-96a0-a88e1024c5ff︠
# The origin has negative gaussian curvature
K(0, 0, 0)
︡b1ae27fe-39c0-45ef-b446-fe367bdf7182︡{"stdout": "-4.0"}︡
︠8911d8e7-80f6-46db-9605-97d99806a98b︠
# To have reasonable bounds on the gaussian curvature, we use the fact that the curvature is most negative at the origin, and goes to zero as x, y go infinity.
c = ColormapTransform(cmsel1, fun=K, bounds=(-4, 0))
︡55bcea4a-007b-4f93-a07e-7781bf439289︡︡
︠9dadc67e-fd4c-42d7-acbe-e8675109e828︠
# Again we see that the curvature falls off pretty quickly away from the origin
c.apply(p)
︡0b462cc9-2913-4a9d-b7cc-b95afc811ddb︡︡
︠7646ed52-b778-47ee-a16f-2e46f53dbbfe︠
# A more complicated example, taken from the implicit_plot3d docs
︡2e710a12-20ea-447d-8299-a4c67282ff12︡︡
︠57970977-0343-4790-885a-b5b4f19986a6︠
T = RDF(golden_ratio)
f = 2 - (cos(x + T*y) + cos(x - T*y) + cos(y + T*z) + cos(y - T*z) + cos(z - T*x) + cos(z + T*x)) 
r = 4.77
p = implicit_plot3d(f, (x, -r, r), (y, -r, r), (z, -r, r), plot_points=40); p
︡d74f6d74-50e6-458a-9ce5-d2ab63d4a42b︡︡
︠007e51c2-9050-446e-8144-ab267cca5e1c︠
K = gaussian_curvature_implicit_surface(f) # Very slow!
︡42fac512-afa8-4024-b5cf-0213f165fc0d︡︡
︠e9df595b-e0a6-4d20-8cb2-e00dcecacd5b︠
c = ColormapTransform(cmsel1, fun=K)
c.apply(p)
︡85665203-d5fb-4d81-b621-e5542e89b87a︡︡
